@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2011 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2012 -- leonerd@leonerd.org.uk
 
 package IO::Socket::IP;
 
@@ -9,36 +9,28 @@ use strict;
 use warnings;
 use base qw( IO::Socket );
 
-our $VERSION = '0.07';
+our $VERSION = '0.16';
 
 use Carp;
 
-BEGIN {
-   # Perl 5.13.9 or above has Socket::getaddrinfo support in core.
-   # Before that we need to use Socket::GetAddrInfo
-   my @imports = qw(
-      getaddrinfo getnameinfo
-      AI_PASSIVE
-      NI_NUMERICHOST NI_NUMERICSERV
-      NI_DGRAM
-   );
-
-   if( require Socket and defined &Socket::getaddrinfo ) {
-      Socket->import( @imports );
-   }
-   else {
-      require Socket::GetAddrInfo;
-      Socket::GetAddrInfo->import( ':newapi', @imports );
-   }
-}
-
-use Socket qw(
-   SOCK_DGRAM SOCK_STREAM IPPROTO_TCP
-   SOL_SOCKET
+use Socket 1.97 qw(
+   getaddrinfo getnameinfo
+   sockaddr_family
+   AF_INET
+   AI_PASSIVE
+   IPPROTO_TCP IPPROTO_UDP
+   IPPROTO_IPV6 IPV6_V6ONLY
+   NI_DGRAM NI_NUMERICHOST NI_NUMERICSERV NIx_NOHOST NIx_NOSERV
    SO_REUSEADDR SO_REUSEPORT SO_BROADCAST SO_ERROR
+   SOCK_DGRAM SOCK_STREAM 
+   SOL_SOCKET
 );
+my $AF_INET6 = eval { Socket::AF_INET6() }; # may not be defined
+my $AI_ADDRCONFIG = eval { Socket::AI_ADDRCONFIG() } || 0;
 use POSIX qw( dup2 );
-use Errno qw( EINPROGRESS );
+use Errno qw( EINVAL EINPROGRESS EISCONN );
+
+use constant HAVE_MSWIN32 => ( $^O eq "MSWin32" );
 
 my $IPv6_re = do {
    # translation of RFC 3986 3.2.2 ABNF to re
@@ -79,8 +71,8 @@ both IPv4 and IPv6
     Type     => SOCK_STREAM,
  ) or die "Cannot construct socket - $@";
 
- my $familyname = ( $sock->sockdomain == AF_INET6 ) ? "IPv6" :
-                  ( $sock->sockdomain == AF_INET  ) ? "IPv4" :
+ my $familyname = ( $sock->sockdomain == PF_INET6 ) ? "IPv6" :
+                  ( $sock->sockdomain == PF_INET  ) ? "IPv4" :
                                                       "unknown";
 
  printf "Connected to google via %s\n", $familyname;
@@ -101,18 +93,18 @@ falling back to IPv4-only on systems which don't.
 
 By placing C<-register> in the import list, C<IO::Socket> uses
 C<IO::Socket::IP> rather than C<IO::Socket::INET> as the class that handles
-C<AF_INET>.  C<IO::Socket> will also use C<IO::Socket::IP> rather than
-C<IO::Socket::INET6> to handle C<AF_INET6>, provided that the C<AF_INET6>
+C<PF_INET>.  C<IO::Socket> will also use C<IO::Socket::IP> rather than
+C<IO::Socket::INET6> to handle C<PF_INET6>, provided that the C<AF_INET6>
 constant is available.
 
 Changing C<IO::Socket>'s default behaviour means that calling the
-C<IO::Socket> constructor with either C<AF_INET> or C<AF_INET6> as the
+C<IO::Socket> constructor with either C<PF_INET> or C<PF_INET6> as the
 C<Domain> parameter will yield an C<IO::Socket::IP> object.
 
  use IO::Socket::IP -register;
 
  my $sock = IO::Socket->new(
-    Domain    => AF_INET6,
+    Domain    => PF_INET6,
     LocalHost => "::1",
     Listen    => 1,
  ) or die "Cannot create socket - $@\n";
@@ -132,10 +124,7 @@ sub import
 
    foreach ( @_ ) {
       if( $_ eq "-register" ) {
-         $pkg->register_domain( Socket::AF_INET() );
-
-         my $AF_INET6 = eval { Socket::AF_INET6() } ||
-                        eval { require Socket6; Socket6::AF_INET6() };
+         $pkg->register_domain( AF_INET );
          $pkg->register_domain( $AF_INET6 ) if defined $AF_INET6;
       }
       else {
@@ -147,6 +136,28 @@ sub import
    goto &IO::Socket::import;
 }
 
+# Convenient capability test function
+{
+   my $can_disable_v6only;
+   sub CAN_DISABLE_V6ONLY
+   {
+      return $can_disable_v6only if defined $can_disable_v6only;
+
+      socket my $testsock, Socket::PF_INET6(), SOCK_STREAM, 0 or
+         die "Cannot socket(PF_INET6) - $!";
+
+      if( setsockopt $testsock, IPPROTO_IPV6, IPV6_V6ONLY, 0 ) {
+         return $can_disable_v6only = 1;
+      }
+      elsif( $! == EINVAL ) {
+         return $can_disable_v6only = 0;
+      }
+      else {
+         die "Cannot setsockopt() - $!";
+      }
+   }
+}
+
 =head1 CONSTRUCTORS
 
 =cut
@@ -176,7 +187,7 @@ C<PeerService> respectively.
 =item PeerAddrInfo => ARRAY
 
 Alternate form of specifying the peer to C<connect()> to. This should be an
-array of the form returned by C<Socket::GetAddrInfo::getaddrinfo>.
+array of the form returned by C<Socket::getaddrinfo>.
 
 This parameter takes precedence over the C<Peer*>, C<Family>, C<Type> and
 C<Proto> arguments.
@@ -198,16 +209,16 @@ C<LocalService> respectively.
 =item LocalAddrInfo => ARRAY
 
 Alternate form of specifying the local address to C<bind()> to. This should be
-an array of the form returned by C<Socket::GetAddrInfo::getaddrinfo>.
+an array of the form returned by C<Socket::getaddrinfo>.
 
 This parameter takes precedence over the C<Local*>, C<Family>, C<Type> and
 C<Proto> arguments.
 
 =item Family => INT
 
-The socket family to pass to C<getaddrinfo> (e.g. C<AF_INET>, C<AF_INET6>).
+The address family to pass to C<getaddrinfo> (e.g. C<AF_INET>, C<AF_INET6>).
 Normally this will be left undefined, and C<getaddrinfo> will search using any
-family supported by the system.
+address family supported by the system.
 
 =item Type => INT
 
@@ -222,6 +233,13 @@ C<'udp'>,C<IPPROTO_UDP>). Normally this will be left undefined, and either
 C<getaddrinfo> or the kernel will choose an appropriate value. May be given
 either in string name or numeric form.
 
+=item GetAddrInfoFlags => INT
+
+More flags to pass to the C<getaddrinfo()> function. These flags will be
+combined with C<AI_ADDRCONFIG>, and if the C<Listen> argument is given,
+C<AI_PASSIVE>. For more information see the documentation about
+C<getaddrinfo()> in the L<Socket> module.
+
 =item Listen => INT
 
 If defined, puts the socket into listening mode where new connections can be
@@ -240,6 +258,32 @@ If true, set the C<SO_REUSEPORT> sockopt (not all OSes implement this sockopt)
 
 If true, set the C<SO_BROADCAST> sockopt
 
+=item V6Only => BOOL
+
+If defined, set the C<IPV6_V6ONLY> sockopt when creating C<PF_INET6> sockets
+to the given value. If true, a listening-mode socket will only listen on the
+C<AF_INET6> addresses; if false it will also accept connections from
+C<AF_INET> addresses.
+
+If not defined, the socket option will not be changed, and default value set
+by the operating system will apply. For repeatable behaviour across platforms
+it is recommended this value always be defined for listening-mode sockets.
+
+Note that not all platforms support disabling this option. Some, at least
+OpenBSD and MirBSD, will fail with C<EINVAL> if you attempt to disable it.
+To determine whether it is possible to disable, you may use the class method
+
+ if( IO::Socket::IP->CAN_DISABLE_V6ONLY ) {
+    ...
+ }
+ else {
+    ...
+ }
+
+If your platform does not support disabling this option but you still want to
+listen for both C<AF_INET> and C<AF_INET6> connections you will have to create
+two listening sockets, one bound to each protocol.
+
 =item Timeout
 
 This C<IO::Socket::INET>-style argument is not currently supported. See the
@@ -247,9 +291,10 @@ C<IO::Socket::INET> INCOMPATIBILITES section below.
 
 =item MultiHomed
 
-This C<IO::Socket::INET>-style argument is not currently supported. See the
-C<IO::Socket::INET> INCOMPATIBILITES section below. However, the behaviour it
-enables is always performed by C<IO::Socket::IP>.
+This C<IO::Socket::INET>-style argument is ignored, except if it is defined
+but false. See the C<IO::Socket::INET> INCOMPATIBILITES section below. 
+
+However, the behaviour it enables is always performed by C<IO::Socket::IP>.
 
 =item Blocking => BOOL
 
@@ -261,7 +306,8 @@ detail.
 
 If neither C<Type> nor C<Proto> hints are provided, a default of
 C<SOCK_STREAM> and C<IPPROTO_TCP> respectively will be set, to maintain
-compatibilty with C<IO::Socket::INET>.
+compatibility with C<IO::Socket::INET>. Other named arguments that are not
+recognised are ignored.
 
 If the constructor fails, it will set C<$@> to an appropriate error message;
 this may be from C<$!> or it may be some other string; not every failure
@@ -307,12 +353,9 @@ sub configure
       my $service = $type . 'Service';
 
       if (exists $arg->{$host} && !exists $arg->{$service}) {
-         local $_ = $arg->{$host};
-         defined or next;
-         if (/\A\[($IPv6_re)\](?::([^\s:]*))?\z/o || /\A([^\s:]*):([^\s:]*)\z/) {
-            $arg->{$host}    = $1;
-            $arg->{$service} = $2 if defined $2 && length $2;
-         }
+         defined $arg->{$host} or next;
+         ( $arg->{$host}, my $s ) = $self->split_addr( $arg->{$host} );
+         $arg->{$service} = $s if defined $s;
       }
    }
 
@@ -328,21 +371,17 @@ sub _configure
    my @localinfos;
    my @peerinfos;
 
-   my @sockopts_enabled;
+   $hints{flags} = ( $arg->{GetAddrInfoFlags} || 0 ) | $AI_ADDRCONFIG;
 
-   if( defined $arg->{Family} ) {
-      my $family = delete $arg->{Family};
+   if( defined( my $family = $arg->{Family} ) ) {
       $hints{family} = $family;
    }
 
-   if( defined $arg->{Type} ) {
-      my $type = delete $arg->{Type};
+   if( defined( my $type = $arg->{Type} ) ) {
       $hints{socktype} = $type;
    }
 
-   if( defined $arg->{Proto} ) {
-      my $proto = delete $arg->{Proto};
-
+   if( defined( my $proto = $arg->{Proto} ) ) {
       unless( $proto =~ m/^\d+$/ ) {
          my $protonum = getprotobyname( $proto );
          defined $protonum or croak "Unrecognised protocol $proto";
@@ -359,14 +398,23 @@ sub _configure
       $hints{protocol} = IPPROTO_TCP;
    }
 
-   if( my $info = delete $arg->{LocalAddrInfo} ) {
+   # Some OSes (NetBSD) don't seem to like just a protocol hint without a
+   # socktype hint as well. We'll set a couple of common ones
+   if( !defined $hints{socktype} and defined $hints{protocol} ) {
+      $hints{socktype} = SOCK_STREAM if $hints{protocol} == IPPROTO_TCP;
+      $hints{socktype} = SOCK_DGRAM  if $hints{protocol} == IPPROTO_UDP;
+   }
+
+   if( my $info = $arg->{LocalAddrInfo} ) {
+      ref $info eq "ARRAY" or croak "Expected 'LocalAddrInfo' to be an ARRAY ref";
       @localinfos = @$info;
    }
    elsif( defined $arg->{LocalHost} or defined $arg->{LocalService} ) {
       # Either may be undef
-      my $host = delete $arg->{LocalHost};
-      my $service = delete $arg->{LocalService};
+      my $host = $arg->{LocalHost};
+      my $service = $arg->{LocalService};
 
+      local $1; # Placate a taint-related bug; [perl #67962]
       defined $service and $service =~ s/\((\d+)\)$// and
          my $fallback_port = $1;
 
@@ -378,18 +426,24 @@ sub _configure
          ( $err, @localinfos ) = getaddrinfo( $host, $fallback_port, \%localhints );
       }
 
-      $err and ( $@ = "$err", return );
+      if( $err ) {
+         $@ = "$err";
+         $! = EINVAL;
+         return;
+      }
    }
 
-   if( my $info = delete $arg->{PeerAddrInfo} ) {
+   if( my $info = $arg->{PeerAddrInfo} ) {
+      ref $info eq "ARRAY" or croak "Expected 'PeerAddrInfo' to be an ARRAY ref";
       @peerinfos = @$info;
    }
    elsif( defined $arg->{PeerHost} or defined $arg->{PeerService} ) {
-      defined( my $host = delete $arg->{PeerHost} ) or
+      defined( my $host = $arg->{PeerHost} ) or
          croak "Expected 'PeerHost'";
-      defined( my $service = delete $arg->{PeerService} ) or
+      defined( my $service = $arg->{PeerService} ) or
          croak "Expected 'PeerService'";
 
+      local $1; # Placate a taint-related bug; [perl #67962]
       defined $service and $service =~ s/\((\d+)\)$// and
          my $fallback_port = $1;
 
@@ -399,21 +453,33 @@ sub _configure
          ( $err, @peerinfos ) = getaddrinfo( $host, $fallback_port, \%hints );
       }
 
-      $err and ( $@ = "$err", return );
+      if( $err ) {
+         $@ = "$err";
+         $! = EINVAL;
+         return;
+      }
    }
 
-   push @sockopts_enabled, SO_REUSEADDR if delete $arg->{ReuseAddr};
-   push @sockopts_enabled, SO_REUSEPORT if delete $arg->{ReusePort};
-   push @sockopts_enabled, SO_BROADCAST if delete $arg->{Broadcast};
+   my @sockopts_enabled;
+   push @sockopts_enabled, SO_REUSEADDR if $arg->{ReuseAddr};
+   push @sockopts_enabled, SO_REUSEPORT if $arg->{ReusePort};
+   push @sockopts_enabled, SO_BROADCAST if $arg->{Broadcast};
 
-   my $listenqueue = delete $arg->{Listen};
+   my $listenqueue = $arg->{Listen};
 
    croak "Cannot Listen with a PeerHost" if defined $listenqueue and @peerinfos;
 
-   my $blocking = delete $arg->{Blocking};
+   my $blocking = $arg->{Blocking};
    defined $blocking or $blocking = 1;
 
-   keys %$arg and croak "Unexpected keys - " . join( ", ", sort keys %$arg );
+   my $v6only = $arg->{V6Only};
+
+   # IO::Socket::INET defines this key. IO::Socket::IP always implements the
+   # behaviour it requests, so we can ignore it, unless the caller is for some
+   # reason asking to disable it.
+   if( defined $arg->{MultiHomed} and !$arg->{MultiHomed} ) {
+      croak "Cannot disable the MultiHomed parameter";
+   }
 
    my @infos;
    foreach my $local ( @localinfos ? @localinfos : {} ) {
@@ -427,7 +493,7 @@ sub _configure
 
          my $family   = $local->{family}   || $peer->{family}   or next;
          my $socktype = $local->{socktype} || $peer->{socktype} or next;
-         my $protocol = $local->{protocol} || $peer->{protocol};
+         my $protocol = $local->{protocol} || $peer->{protocol} || 0;
 
          push @infos, {
             family    => $family,
@@ -448,14 +514,15 @@ sub _configure
    ${*$self}{io_socket_ip_idx} = -1;
 
    ${*$self}{io_socket_ip_sockopts} = \@sockopts_enabled;
+   ${*$self}{io_socket_ip_v6only} = $v6only;
    ${*$self}{io_socket_ip_listenqueue} = $listenqueue;
    ${*$self}{io_socket_ip_blocking} = $blocking;
 
    ${*$self}{io_socket_ip_errors} = [ undef, undef, undef ];
 
-   if( $blocking ) {
-      $self->setup or return undef;
-   }
+   # ->setup is allowed to return false in nonblocking mode
+   $self->setup or !$blocking or return undef;
+
    return $self;
 }
 
@@ -478,6 +545,11 @@ sub setup
          $self->setsockopt( SOL_SOCKET, $sockopt, pack "i", 1 ) or ( $@ = "$!", return undef );
       }
 
+      if( defined ${*$self}{io_socket_ip_v6only} and defined $AF_INET6 and $info->{family} == $AF_INET6 ) {
+         my $v6only = ${*$self}{io_socket_ip_v6only};
+         $self->setsockopt( IPPROTO_IPV6, IPV6_V6ONLY, pack "i", $v6only ) or ( $@ = "$!", return undef );
+      }
+
       if( defined( my $addr = $info->{localaddr} ) ) {
          $self->bind( $addr ) or
             ( ${*$self}{io_socket_ip_errors}[1] = $!, next );
@@ -488,16 +560,15 @@ sub setup
       }
 
       if( defined( my $addr = $info->{peeraddr} ) ) {
-         # It seems that IO::Socket hides EINPROGRESS errors, making them look
-         # like a success. This is annoying here.
          if( $self->connect( $addr ) ) {
-            $! = EINPROGRESS, return 0 if !$self->connected; # EINPROGRESS
-
             $! = 0;
             return 1;
          }
 
-         return 0 if $! == EINPROGRESS;
+         if( $! == EINPROGRESS or HAVE_MSWIN32 && $! == Errno::EWOULDBLOCK() ) {
+            ${*$self}{io_socket_ip_connect_in_progress} = 1;
+            return 0;
+         }
 
          ${*$self}{io_socket_ip_errors}[0] = $!;
          next;
@@ -506,8 +577,6 @@ sub setup
       return 1;
    }
 
-   $self->close;
-
    # Pick the most appropriate error, stringified
    $! = ( grep defined, @{ ${*$self}{io_socket_ip_errors}} )[0];
    $@ = "$!";
@@ -517,16 +586,47 @@ sub setup
 sub connect
 {
    my $self = shift;
-   return $self->SUPER::connect( @_ ) if @_;
 
-   $! = 0, return 1 if $self->fileno and defined $self->peername;
+   # It seems that IO::Socket hides EINPROGRESS errors, making them look like
+   # a success. This is annoying here.
+   # Instead of putting up with its frankly-irritating intentional breakage of
+   # useful APIs I'm just going to end-run around it and call CORE::connect()
+   # directly
 
-   if( $self->fileno ) {
-      # A connect has just failed, get its error value
-      ${*$self}{io_socket_ip_errors}[0] = $self->getsockopt( SOL_SOCKET, SO_ERROR );
+   return CORE::connect( $self, $_[0] ) if @_;
+
+   return 1 if !${*$self}{io_socket_ip_connect_in_progress};
+
+   # See if a connect attempt has just failed with an error
+   if( my $errno = $self->getsockopt( SOL_SOCKET, SO_ERROR ) ) {
+      delete ${*$self}{io_socket_ip_connect_in_progress};
+      ${*$self}{io_socket_ip_errors}[0] = $! = $errno;
+      return $self->setup;
    }
 
-   return $self->setup;
+   # No error, so either connect is still in progress, or has completed
+   # successfully. We can tell by trying to connect() again; either it will
+   # succeed or we'll get EISCONN (connected successfully), or EALREADY
+   # (still in progress). This even works on MSWin32.
+   my $addr = ${*$self}{io_socket_ip_infos}[${*$self}{io_socket_ip_idx}]{peeraddr};
+
+   if( $self->connect( $addr ) or $! == EISCONN ) {
+      delete ${*$self}{io_socket_ip_connect_in_progress};
+      $! = 0;
+      return 1;
+   }
+   else {
+      $! = EINPROGRESS;
+      return 0;
+   }
+}
+
+sub connected
+{
+   my $self = shift;
+   return defined $self->fileno &&
+          !${*$self}{io_socket_ip_connect_in_progress} &&
+          defined $self->peername;
 }
 
 =head1 METHODS
@@ -539,19 +639,32 @@ L<IO::Socket> and L<IO::Handle>.
 sub _get_host_service
 {
    my $self = shift;
-   my ( $addr, $numeric ) = @_;
-
-   my $flags = 0;
+   my ( $addr, $flags, $xflags ) = @_;
 
    $flags |= NI_DGRAM if $self->socktype == SOCK_DGRAM;
-   $flags |= NI_NUMERICHOST|NI_NUMERICSERV if $numeric;
 
-   my ( $err, $host, $service ) = getnameinfo( $addr, $flags );
+   my ( $err, $host, $service ) = getnameinfo( $addr, $flags, $xflags || 0 );
    croak "getnameinfo - $err" if $err;
 
    return ( $host, $service );
 }
 
+sub _unpack_sockaddr
+{
+   my ( $addr ) = @_;
+   my $family = sockaddr_family $addr;
+
+   if( $family == AF_INET ) {
+      return ( Socket::unpack_sockaddr_in( $addr ) )[1];
+   }
+   elsif( defined $AF_INET6 and $family == $AF_INET6 ) {
+      return ( Socket::unpack_sockaddr_in6( $addr ) )[1];
+   }
+   else {
+      croak "Unrecognised address family $family";
+   }
+}
+
 =head2 ( $host, $service ) = $sock->sockhost_service( $numeric )
 
 Returns the hostname and service name of the local address (that is, the
@@ -562,7 +675,7 @@ resolved into names.
 
 The following four convenience wrappers may be used to obtain one of the two
 values returned here. If both host and service names are required, this method
-is preferrable to the following wrappers, because it will call
+is preferable to the following wrappers, because it will call
 C<getnameinfo(3)> only once.
 
 =cut
@@ -572,12 +685,12 @@ sub sockhost_service
    my $self = shift;
    my ( $numeric ) = @_;
 
-   $self->_get_host_service( $self->sockname, $numeric );
+   $self->_get_host_service( $self->sockname, $numeric ? NI_NUMERICHOST|NI_NUMERICSERV : 0 );
 }
 
 =head2 $addr = $sock->sockhost
 
-Return the numeric form of the local address
+Return the numeric form of the local address as a textual representation
 
 =head2 $port = $sock->sockport
 
@@ -593,11 +706,19 @@ Return the resolved name of the local port number
 
 =cut
 
-sub sockhost { ( shift->sockhost_service(1) )[0] }
-sub sockport { ( shift->sockhost_service(1) )[1] }
+sub sockhost { my $self = shift; ( $self->_get_host_service( $self->sockname, NI_NUMERICHOST, NIx_NOSERV ) )[0] }
+sub sockport { my $self = shift; ( $self->_get_host_service( $self->sockname, NI_NUMERICSERV, NIx_NOHOST ) )[1] }
 
-sub sockhostname { ( shift->sockhost_service(0) )[0] }
-sub sockservice  { ( shift->sockhost_service(0) )[1] }
+sub sockhostname { my $self = shift; ( $self->_get_host_service( $self->sockname, 0, NIx_NOSERV ) )[0] }
+sub sockservice  { my $self = shift; ( $self->_get_host_service( $self->sockname, 0, NIx_NOHOST ) )[1] }
+
+=head2 $addr = $sock->sockaddr
+
+Return the local address as a binary octet string
+
+=cut
+
+sub sockaddr { my $self = shift; _unpack_sockaddr $self->sockname }
 
 =head2 ( $host, $service ) = $sock->peerhost_service( $numeric )
 
@@ -607,7 +728,7 @@ C<sockhost_service> method.
 
 The following four convenience wrappers may be used to obtain one of the two
 values returned here. If both host and service names are required, this method
-is preferrable to the following wrappers, because it will call
+is preferable to the following wrappers, because it will call
 C<getnameinfo(3)> only once.
 
 =cut
@@ -617,12 +738,12 @@ sub peerhost_service
    my $self = shift;
    my ( $numeric ) = @_;
 
-   $self->_get_host_service( $self->peername, $numeric );
+   $self->_get_host_service( $self->peername, $numeric ? NI_NUMERICHOST|NI_NUMERICSERV : 0 );
 }
 
 =head2 $addr = $sock->peerhost
 
-Return the numeric form of the peer address
+Return the numeric form of the peer address as a textual representation
 
 =head2 $port = $sock->peerport
 
@@ -638,11 +759,19 @@ Return the resolved name of the peer port number
 
 =cut
 
-sub peerhost    { ( shift->peerhost_service(1) )[0] }
-sub peerport    { ( shift->peerhost_service(1) )[1] }
+sub peerhost { my $self = shift; ( $self->_get_host_service( $self->peername, NI_NUMERICHOST, NIx_NOSERV ) )[0] }
+sub peerport { my $self = shift; ( $self->_get_host_service( $self->peername, NI_NUMERICSERV, NIx_NOHOST ) )[1] }
 
-sub peerhostname { ( shift->peerhost_service(0) )[0] }
-sub peerservice  { ( shift->peerhost_service(0) )[1] }
+sub peerhostname { my $self = shift; ( $self->_get_host_service( $self->peername, 0, NIx_NOSERV ) )[0] }
+sub peerservice  { my $self = shift; ( $self->_get_host_service( $self->peername, 0, NIx_NOHOST ) )[1] }
+
+=head2 $addr = $peer->peeraddr
+
+Return the peer address as a binary octet string
+
+=cut
+
+sub peeraddr { my $self = shift; _unpack_sockaddr $self->peername }
 
 # This unbelievably dodgy hack works around the bug that IO::Socket doesn't do
 # it
@@ -685,11 +814,13 @@ non-blocking connect.
 When using non-blocking mode, the caller must repeatedly check for
 writeability on the filehandle (for instance using C<select> or C<IO::Poll>).
 Each time the filehandle is ready to write, the C<connect> method must be
-called, with no arguments.
+called, with no arguments. Note that some operating systems, most notably
+C<MSWin32> do not report a C<connect()> failure using write-ready; so you must
+also C<select()> for exceptional status.
 
 While C<connect> returns false, the value of C<$!> indicates whether it should
-be tried again (by being set to the value C<EINPROGRESS>), or whether a
-permanent error has occured (e.g. C<ECONNREFUSED>).
+be tried again (by being set to the value C<EINPROGRESS>, or C<EWOULDBLOCK> on
+MSWin32), or whether a permanent error has occurred (e.g. C<ECONNREFUSED>).
 
 Once the socket has been connected to the peer, C<connect> will return true
 and the socket will now be ready to use.
@@ -704,7 +835,7 @@ achieved by using L<Net::LibAsyncNS>, or the C<getaddrinfo(3)> function can be
 called in a child process.
 
  use IO::Socket::IP;
- use Errno qw( EINPROGRESS );
+ use Errno qw( EINPROGRESS EWOULDBLOCK );
 
  my @peeraddrinfo = ... # Caller must obtain the getaddinfo result here
 
@@ -713,11 +844,13 @@ called in a child process.
     Blocking     => 0,
  ) or die "Cannot construct socket - $@";
 
- while( !$socket->connect and $! == EINPROGRESS ) {
+ while( !$socket->connect and ( $! == EINPROGRESS || $! == EWOULDBLOCK ) ) {
     my $wvec = '';
     vec( $wvec, fileno $socket, 1 ) = 1;
+    my $evec = '';
+    vec( $evec, fileno $socket, 1 ) = 1;
 
-    select( undef, $wvec, undef, undef ) or die "Cannot select - $!";
+    select( undef, $wvec, $evec, undef ) or die "Cannot select - $!";
  }
 
  die "Cannot connect - $!" if $!;
@@ -734,6 +867,8 @@ C<PF_INET6>.
 For another example using C<IO::Poll> and C<Net::LibAsyncNS>, see the
 F<examples/nonblocking_libasyncns.pl> file in the module distribution.
 
+=cut
+
 =head1 C<PeerHost> AND C<LocalHost> PARSING
 
 To support the C<IO::Socket::INET> API, the host and port information may be
@@ -747,8 +882,8 @@ The value of the C<...Host> argument will be split to give both the hostname
 and port (or service name):
 
  hostname.example.org:http    # Host name
- 10.0.0.1:80                  # IPv4 address
- [fe80::123]:443              # IPv6 address
+ 192.0.2.1:80                 # IPv4 address
+ [2001:db8::1]:80             # IPv6 address
 
 In each case, the port or service name (e.g. C<80>) is passed as the
 C<LocalService> or C<PeerService> argument.
@@ -762,18 +897,56 @@ service name and number, in a form such as
 In this case, the name (C<http>) will be tried first, but if the resolver does
 not understand it then the port number (C<80>) will be used instead.
 
+=head2 ( $host, $port ) = IO::Socket::IP->split_addr( $addr )
+
+Utility method that provides the parsing functionality described above.
+Returns a 2-element list, containing either the split hostname and port
+description if it could be parsed, or the given address and C<undef> if it was
+not recognised.
+
+ IO::Socket::IP->split_addr( "hostname:http" )
+                              # ( "hostname",  "http" )
+
+ IO::Socket::IP->split_addr( "192.0.2.1:80" )
+                              # ( "192.0.2.1", "80"   )
+
+ IO::Socket::IP->split_addr( "[2001:db8::1]:80" )
+                              # ( "2001:db8::1", "80" )
+
+ IO::Socket::IP->split_addr( "something.else" )
+                              # ( "something.else", undef )
+
+=cut
+
+sub split_addr
+{
+   shift;
+   my ( $addr ) = @_;
+
+   local ( $1, $2 ); # Placate a taint-related bug; [perl #67962]
+   if( $addr =~ m/\A\[($IPv6_re)\](?::([^\s:]*))?\z/ or
+       $addr =~ m/\A([^\s:]*):([^\s:]*)\z/ ) {
+      return ( $1, $2 ) if defined $2 and length $2;
+      return ( $1, undef );
+   }
+
+   return ( $addr, undef );
+}
+
 =head1 C<IO::Socket::INET> INCOMPATIBILITES
 
 =over 4
 
 =item *
 
-The C<Timeout> and C<MultiHomed> constructor arguments are currently not
-recognised.
+The C<Timeout> constructor argument is currently not recognised.
 
 The behaviour enabled by C<MultiHomed> is in fact implemented by
 C<IO::Socket::IP> as it is required to correctly support searching for a
-useable address from the results of the C<getaddrinfo(3)> call.
+useable address from the results of the C<getaddrinfo(3)> call. The
+constructor will ignore the value of this argument, except if it is defined
+but false. An exception is thrown in this case, because that would request it
+disable the C<getaddrinfo(3)> search behaviour in the first place.
 
 =back
 
@@ -785,34 +958,11 @@ useable address from the results of the C<getaddrinfo(3)> call.
 
 =item *
 
-Cache the returns from C<sockhost_service> and C<peerhost_service> to avoid
-double-lookup overhead in such code as
-
-  printf "Peer is %s:%d\n", $sock->peerhost, $sock->peerport;
-
-=item *
-
 Investigate whether C<POSIX::dup2> upsets BSD's C<kqueue> watchers, and if so,
 consider what possible workarounds might be applied.
 
 =back
 
-=head1 BUGS
-
-=over 4
-
-=item *
-
-Nonblocking connect fails unit tests on MSWin32 smoke-testing machines. The
-specifics of the failure are that C<connect()> seems to block anyway despite
-being asked not to, and that failure to connect is not detected properly. I am
-as yet unsure why this is.
-
-Blocking connect on MSWin32, and both blocking and nonblocking connect on
-other platforms, all test OK on smoke testing.
-
-=back
-
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
